
/*################################################################################################
 *##################                                                            ##################
 *##################                     Execution module                       ##################
 *##################                                                            ##################
 *##################                      Implementation                        ##################
 *##################                                                            ##################
 *################################################################################################
 */

/*
 * OKit:   Object Tree,  RPL Virtual Machine  and  compiler  shared library
 *
 * Copyright (C) 2000 Yann LANDRIN-SCHWEITZER
 *
 *
 * This library is free software;  you can redistribute it and/or modify it
 * under  the terms of the  GNU Lesser General Public License  as published
 * by the  Free Software Foundation;  either version 2.1 of the License, or
 * (at your option) any later version.
 *
 * This  library  is  distributed  in  the hope that it will be useful, but
 * WITHOUT   ANY   WARRANTY;   without   even   the   implied  warranty  of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library;  if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 *
 * Author:   Yann LANDRIN-SCHWEITZER
 * Contact:  landrin@multimania.com
 * Homepage: http://landrin.multimania.com/
 */



#include <stdlib.h>
#include <unistd.h>
#include <errno.h>

#include <OKit/ErrLog.h>

#include <OKit/SysTypes.h>
#include <OKit/ObjTypes.h>

#include <OKit/ExecMod.h>



//////////////////////////////////////////////////////////////////////////////////////////////////
/////
/////  Standard error messages
/////


char * ExecError[] = {

	"",					//  0 OK
	"Unhandled error",			//  1 UNHDLD
	"Run time error",			//  2 RNTERR
	"Incoherent data",			//  3 DTINCR
	"System error",				//  4 SYSERR
	"IO error",				//  5 IOSERR
	"Exec error",				//  6 EXCERR
	"Unknown built-in operator",		//  7 NOBLTN
	"Not implemented",			//  8 UNIMPL
	"Unknown user-defined operator",	//  9 NOUSER

	"Too few arguments",			// 10 TOOFEWARG
	"Invalid argument type",		// 11 INVARGTYP
	"Invalid argument value",		// 12 INVARGVAL
	"",					// 13 
	"",					// 14 
	"",					// 15 
	"",					// 16 
	"",					// 17 
	"",					// 18 
	"",					// 19 
} ;


//////////////////////////////////////////////////////////////////////////////////////////////////
/////
/////  Object evaluation
/////


inline Status _ObjectEval(Stack stk, NameSpace dom, Object obj) ;
inline Status _ListEval(Stack stk, NameSpace dom, List lst) ;
inline Status _CodeEval(Stack stk, NameSpace dom, Code cod) ;
inline Status _BltnEval(Stack stk, NameSpace dom, Bltn blt) ;


inline Status _ObjectEval(Stack stk, NameSpace dom, Object obj) {
	Status erst = {-1, "<!>", ""} ;
	Status okst = { 0,    "", ""} ;
	if(obj==NULL) {
		errno = EINVAL ;
		DEBUG("ObjectEval",1,"NULL Object") ;
		return erst ;
	}
	switch(obj->type) {
		case OT_VOID:
			if(StackPush(stk,obj)==-1) {
				return erst ;
			}
			return okst ;
		case OT_INTG:
			if(StackPush(stk,obj)==-1) {
				return erst ;
			}
			return okst ;
		case OT_FLTP:
			if(StackPush(stk,obj)==-1) {
				return erst ;
			}
			return okst ;
		case OT_STRG:
			if(StackPush(stk,obj)==-1) {
				return erst ;
			}
			return okst ;
		case OT_LIST:
			return _ListEval(stk,dom,(List)obj->obj) ;
		case OT_CODE:
			return _CodeEval(stk,dom,(Code)obj->obj) ;
		case OT_BLTN:
			return _BltnEval(stk,dom,(Bltn)obj->obj) ;
	}
	return erst ;
}


inline Status _ListEval(Stack stk, NameSpace dom, List lst) {
	struct S_Node * nod ;
	Status erst = {-1, "<!>", ""} ;
	Status okst = { 0,    "", ""} ;
	Status stat ;
	if(lst==NULL) {
		errno = EINVAL ;
		DEBUG("ListEval",1,"NULL List") ;
		return erst ;
	}
	nod = lst->frst ;
	while(nod!=NULL) {
		if(nod->obj==NULL) {
			errno = EINVAL ;
			DEBUG("ListEval",1,"NULL Object") ;
			return erst ;
		}
		switch(nod->obj->type) {
			case OT_VOID:
				if(StackPush(stk,nod->obj)==-1) {
					return erst ;
				}
				break ;
			case OT_INTG:
				if(StackPush(stk,nod->obj)==-1) {
					return erst ;
				}
				break ;
			case OT_FLTP:
				if(StackPush(stk,nod->obj)==-1) {
					return erst ;
				}
				break ;
			case OT_STRG:
				if(StackPush(stk,nod->obj)==-1) {
					return erst ;
				}
				break ;
			case OT_LIST:
				if(StackPush(stk,nod->obj)==-1) {
					return erst ;
				}
				break ;
			case OT_CODE:
				stat = _CodeEval(stk,dom,(Code)nod->obj->obj) ;
				if(stat.num) {
					return stat ;
				}
				break ;
			case OT_BLTN:
				stat = _BltnEval(stk,dom,(Bltn)nod->obj->obj) ;
				if(stat.num) {
					return stat ;
				}
				break ;
			default:
				return erst ;
		}
		if(nod==lst->last) break ;
		nod = nod->next ;
	}
	return okst ;
}


inline Status _CodeEval(Stack stk, NameSpace dom, Code cod) {
	Status erst = {-1, "<!>", ""} ;
	Status okst = { 0,    "", ""} ;
	Status stat ;
	unsigned long i ;
	if(cod==NULL) {
		errno = EINVAL ;
		DEBUG("CodeEval",1,"NULL Code") ;
		return erst ;
	}
	if(cod->func==NULL) {
		return okst ;
	}
	if(StackDepth(stk)<cod->argc) {
		stat.num = TOOFEWARG ;
		stat.msg = ExecError[TOOFEWARG] ;
		stat.com[0] = 0 ;
		return stat ;
	}
	if(StackChkType(stk,cod->argc,cod->argt)) {
		stat.num = INVARGTYP ;
		stat.msg = ExecError[INVARGTYP] ;
		stat.com[0] = 0 ;
		return stat ;
	}
	stat = ((CodeOperator)cod->func)(stk,dom,cod->data) ;
	if(stat.num) {
		i = 0 ;
		while( (i<STAT_NAME_LEN) && (i<CODE_NAME_LEN) ) {
			if(cod->name[i]==0) break ;
			stat.com[i] = cod->name[i] ;
			i ++ ;
		}
		stat.com[i] = 0 ;
		return stat ;
	}
	return okst ;
}


inline Status _BltnEval(Stack stk, NameSpace dom, Bltn blt) {
	Status erst = {-1, "<!>", ""} ;
	Status okst = { 0,    "", ""} ;
	Status stat ;
	unsigned long i ;
	if(blt==NULL) {
		errno = EINVAL ;
		DEBUG("BltnEval",1,"NULL Bltn") ;
		return erst ;
	}
	if(blt->func==NULL) {
		errno = EINVAL ;
		DEBUG("BltnEval",1,"NULL Bltn Operator") ;
		return erst ;
	}
	if(StackDepth(stk)<blt->argc) {
		stat.num = TOOFEWARG ;
		stat.msg = ExecError[TOOFEWARG] ;
		stat.com[0] = 0 ;
		return stat ;
	}
	if(StackChkType(stk,blt->argc,blt->argt)) {
		stat.num = INVARGTYP ;
		stat.msg = ExecError[INVARGTYP] ;
		stat.com[0] = 0 ;
		return stat ;
	}
	stat = ((BltnOperator)blt->func)(stk,dom) ;
	if(stat.num) {
		i = 0 ;
		while( (i<STAT_NAME_LEN) && (i<BLTN_NAME_LEN) ) {
			if(blt->name[i]==0) break ;
			stat.com[i] = blt->name[i] ;
			i ++ ;
		}
		stat.com[i] = 0 ;
		return stat ;
	}
	return okst ;
}



Status ObjectEval(Stack stk, NameSpace dom, Object obj) {
	Status erst = {-1, "<!>", ""} ;
	if(obj==NULL) {
		errno = EINVAL ;
		DEBUG("ObjectEval",1,"NULL Object") ;
		return erst ;
	}
	if(stk==NULL) {
		errno = EINVAL ;
		DEBUG("ObjectEval",1,"NULL Stack") ;
		return erst ;
	}
	if(dom==NULL) {
		errno = EINVAL ;
		DEBUG("ObjectEval",1,"NULL NameSpace") ;
		return erst ;
	}
	return _ObjectEval(stk,dom,obj) ;
}


Status ListEval(Stack stk, NameSpace dom, List lst) {
	Status erst = {-1, "<!>", ""} ;
	if(stk==NULL) {
		errno = EINVAL ;
		DEBUG("ListEval",1,"NULL Stack") ;
		return erst ;
	}
	if(dom==NULL) {
		errno = EINVAL ;
		DEBUG("ListEval",1,"NULL NameSpace") ;
		return erst ;
	}
	if(lst==NULL) {
		errno = EINVAL ;
		DEBUG("ListEval",1,"NULL List") ;
		return erst ;
	}
	return _ListEval(stk,dom,lst) ;
}


Status CodeEval(Stack stk, NameSpace dom, Code cod) {
	Status erst = {-1, "<!>", ""} ;
	if(cod==NULL) {
		errno = EINVAL ;
		DEBUG("CodeEval",1,"NULL Code") ;
		return erst ;
	}
	if(stk==NULL) {
		errno = EINVAL ;
		DEBUG("CodeEval",1,"NULL Stack") ;
		return erst ;
	}
	if(dom==NULL) {
		errno = EINVAL ;
		DEBUG("CodeEval",1,"NULL NameSpace") ;
		return erst ;
	}
	if(cod->func==NULL) {
		errno = EINVAL ;
		DEBUG("CodeEval",1,"NULL Code Operator") ;
		return erst ;
	}
	return _CodeEval(stk,dom,cod) ;
}


Status BltnEval(Stack stk, NameSpace dom, Bltn blt) {
	Status erst = {-1, "<!>", ""} ;
	if(blt==NULL) {
		errno = EINVAL ;
		DEBUG("BltnEval",1,"NULL Bltn") ;
		return erst ;
	}
	if(stk==NULL) {
		errno = EINVAL ;
		DEBUG("BltnEval",1,"NULL Stack") ;
		return erst ;
	}
	if(dom==NULL) {
		errno = EINVAL ;
		DEBUG("BltnEval",1,"NULL NameSpace") ;
		return erst ;
	}
	if(blt->func==NULL) {
		errno = EINVAL ;
		DEBUG("CodeEval",1,"NULL Bltn Operator") ;
		return erst ;
	}
	return _BltnEval(stk,dom,blt) ;
}





