
/*################################################################################################
 *##################                                                            ##################
 *##################                       System Types                         ##################
 *##################                                                            ##################
 *##################                      Implementation                        ##################
 *##################                                                            ##################
 *################################################################################################
 */

/*
 * OKit:   Object Tree,  RPL Virtual Machine  and  compiler  shared library
 *
 * Copyright (C) 2000 Yann LANDRIN-SCHWEITZER
 *
 *
 * This library is free software;  you can redistribute it and/or modify it
 * under  the terms of the  GNU Lesser General Public License  as published
 * by the  Free Software Foundation;  either version 2.1 of the License, or
 * (at your option) any later version.
 *
 * This  library  is  distributed  in  the hope that it will be useful, but
 * WITHOUT   ANY   WARRANTY;   without   even   the   implied  warranty  of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library;  if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 *
 * Author:   Yann LANDRIN-SCHWEITZER
 * Contact:  landrin@multimania.com
 * Homepage: http://landrin.multimania.com/
 */



#include <stdlib.h>
#include <unistd.h>
#include <stdio.h>
#include <errno.h>

#include <Internal.h>
#include <OKit/ErrLog.h>
#include <OKit/ObjTypes.h>

#include <OKit/SysTypes.h>



//////////////////////////////////////////////////////////////////////////////////////////////////
/////
/////  Stack definition
/////

// struct S_Stack {
//	unsigned long size ;
//	unsigned long chnk ;
//	unsigned long dpth ;
//	Object * data ;
// }
//
// typedef struct S_Stack * Stack ;


Stack StackAlloc(unsigned long chksz) {
	Stack stk ;
	stk = (Stack) malloc(sizeof(struct S_Stack)) ;
	if(stk==NULL) {
		DEBUG("StackAlloc",0,NULL) ;
		return NULL ;
	}
	stk->chnk = (chksz>0)?chksz:DEFCHNKSIZE ;
	stk->dpth = 0 ;
	stk->size = 0 ;
	stk->data = NULL ;
	return stk ;
}


void StackFree(Stack stk) {
	register unsigned long i = 0 ;
	if(stk!=NULL) {
		while(i<stk->dpth) {
			ObjectDrop(stk->data[i]) ;
			ObjectFree(stk->data[i]) ;
			i ++ ;
		}
		free(stk->data) ;
		stk->data = NULL ;
		stk->size = 0 ;
		stk->dpth = 0 ;
		free(stk) ;
	}
}


unsigned long StackDepth(Stack stk) {
	if(stk==NULL) {
		errno = EINVAL ;
		DEBUG("StackDepth",2,"NULL Stack") ;
		return -1 ;
	}
	return stk->dpth ;
}


unsigned long StackPush(Stack stk, Object obj) {
	Object * new ;
	if(stk==NULL) {
		errno = EINVAL ;
		DEBUG("StackPush",2,"NULL Stack") ;
		return -1 ;
	}
	if(obj==NULL) {
		errno = EINVAL ;
		DEBUG("StackPush",2,"NULL Object") ;
		return -1 ;
	}
	if(stk->chnk==0) {
		errno = EACCES ;
		DEBUG("StackPush",3,"Freezed Stack") ;
		return -1 ;
	}
	if(stk->dpth==stk->size) {
		new = (Object *) realloc(stk->data,(stk->size+stk->chnk)*sizeof(struct S_Object)) ;
		if(new==NULL) {
			DEBUG("StackPush",0,NULL) ;
			return -1 ;
		}
		stk->data = new ;
		stk->size += stk->chnk ;
	}
	ObjectDup(obj) ;
	stk->data[stk->dpth] = obj ;
	stk->dpth ++ ;
	return stk->dpth ;
}


Object StackPop(Stack stk) {
	Object obj ;
	Object * new ;
	if(stk==NULL) {
		errno = EINVAL ;
		DEBUG("StackPop",2,"NULL Stack") ;
		return NULL ;
	}
	if(stk->dpth==0) {
		errno = EINVAL ;
		DEBUG("StackPop",2,"Invalid index") ;
		return NULL ;
	}
	if(stk->chnk==0) {
		errno = EACCES ;
		DEBUG("StackPop",3,"Freezed Stack") ;
		return NULL ;
	}
	stk->dpth -- ;
	obj = stk->data[stk->dpth] ;
	ObjectDrop(obj) ;
	if(stk->dpth+2*stk->chnk<stk->size) {
		new = (Object *) realloc(stk->data,(stk->size-stk->chnk)*sizeof(struct S_Object)) ;
		if(new==NULL) { // Non-fatal(!) error: can't disallocate
			DEBUG("StackPop",0,"Disallocation not performed") ;
			return obj ;
		}
		stk->data = new ;
		stk->size -= stk->chnk ;
	}
	return obj ;
}


Object StackGet(Stack stk, unsigned long num) {
	if(stk==NULL) {
		errno = EINVAL ;
		DEBUG("StackGet",2,"NULL Stack") ;
		return NULL ;
	}
	if((num==0)||(num>stk->dpth)) {
		errno = EINVAL ;
		DEBUG("StackGet",2,"Invalid index") ;
		return NULL ;
	}
	return stk->data[stk->dpth-num] ;
}


Object StackPut(Stack stk, Object obj, unsigned long num) {
	Object old ;
	if(stk==NULL) {
		errno = EINVAL ;
		DEBUG("StackPut",2,"NULL Stack") ;
		return NULL ;
	}
	if(obj==NULL) {
		errno = EINVAL ;
		DEBUG("StackPut",2,"NULL Object") ;
		return NULL ;
	}
	if((num==0)||(num>stk->dpth)) {
		errno = EINVAL ;
		DEBUG("StackPut",2,"Invalid index") ;
		return NULL ;
	}
	old = stk->data[stk->dpth-num] ;
	stk->data[stk->dpth-num] = obj ;
	return old ;
}


unsigned long StackChkType(Stack stk, unsigned long argc, Type * argt) {
	register unsigned long i = 0 ;
	register unsigned long f = 0 ;
	if(stk==NULL) {
		errno = EINVAL ;
		DEBUG("StackChkType",2,"NULL Stack") ;
		return -1 ;
	}
	if(stk->dpth<argc) {
		return -1 ;
	}
	while(i<argc) {
		// Bit [i] of f is an error flag. If it is 1, then the object type did
		// not match. This is true when no bits in the i-th pos were 1 both in
		// 'argt' (required type) and in the 'type' field of the object at the
		// corresponding level.
		f = f | ( (( argt[i] & (stk->data[stk->dpth-i-1]->type) )?0:1) <<i ) ;
		i ++ ;
	}
	return f ;
}


unsigned long StackOut(Stack stk, char * out, unsigned long max, char * head, char * rowfmt, char * tail, unsigned long rows) {
	register unsigned long n = 0 ;
	register unsigned long i ;
	register unsigned long j ;
	register unsigned long u ;
	register unsigned long r ;
	unsigned long l ;
	char fil = ' ' ;  //Fill char

	if(stk==NULL) {
		errno = EINVAL ;
		DEBUG("StackOut",2,"NULL Stack") ;
		return -1 ;
	}

	if(out==NULL) {
		j = 0 ;
		while(rowfmt[j]!=0) {
			u = j ;
			while(rowfmt[j]!=0) {
				if(rowfmt[j]=='%') break ;
				j ++ ;
				n ++ ;
			}
			if(rowfmt[j]==0) break ;
			j ++ ;
			l = 0 ;
			if(rowfmt[j]==0) break ;
			if( (rowfmt[j]<='0')||(rowfmt[j]>'9') ) {
				j ++ ;
			}
			if(rowfmt[j]==0) break ;
			if( (rowfmt[j]>='0')&&(rowfmt[j]<='9') ) {
				j += _scaninteger(&l,rowfmt+j,10,10) ;
			}
			if(rowfmt[j]==0) break ;
			switch(rowfmt[j]) {
				case 'n':
					if(l==0) l = 4 ;
					n += l ;
					break ;
				case 's':
					if(l==0) l = 20 ;
					n += l ;
					break ;
				default:
					;
			}
			j ++ ;
		}
		n *= rows ;
		j = 0 ;
		while(head[j]!=0) {
			if(head[j]==0) break ;
			j ++ ;
			n ++ ;
		}
		j = 0 ;
		while(tail[j]!=0) {
			j ++ ;
			n ++ ;
		}
		return n ;
	}

	//First line
	j = 0 ;
	while(n<max) {
		if(head[j]==0) break ;
		out[n] = head[j] ;
		j ++ ;
		n ++ ;
	}
	if(n<max) {
		out[n] = '\n' ;
		n ++ ;
	}
	else {
		out[n] = 0 ;
		return n ;
	}
	i = rows ;
	// Outing empty stack lines
	while(i>stk->dpth) {
		// One empty line more...
		j = 0 ;
		while( (n<max)&&(rowfmt[j]!=0) ) {
			while( (n<max)&&(rowfmt[j]!=0) ) {
				if(rowfmt[j]=='%') break ;
				out[n] = rowfmt[j] ;
				j ++ ;
				n ++ ;
			}
			if(n==max) {
				out[n] = 0 ;
				return n ;
			}
			if(rowfmt[j]==0) break ;
			j ++ ;
			l = 0 ;
			if(rowfmt[j]==0) break ;
			if( (rowfmt[j]<='0')||(rowfmt[j]>'9') ) {
				fil = rowfmt[j] ;
				j ++ ;
			}
			else fil = ' ' ;
			if(rowfmt[j]==0) break ;
			if( (rowfmt[j]>='0')&&(rowfmt[j]<='9') ) {
				j += _scaninteger(&l,rowfmt+j,10,10) ;
			}
			if(rowfmt[j]==0) break ;
			switch(rowfmt[j]) {
				case 'n':
					if(l==0) l = 4 ;
					if(n+l<max) {
						_printinteger(i,out+n,l,10) ;
						n += l ;
					}
					break ;
				case 's':
					if(l==0) l = 20 ;
					if(n+l<max) {
						u = l ;
						while(u>0) {
							u -- ;
							out[n] = fil ;
							n ++ ;
						}
					}
					break ;
				default:
					;
			}
			j ++ ;
		}
		if(n<max) {
			out[n] = '\n' ;
			n ++ ;
		}
		else {
			out[n] = 0 ;
			return n ;
		}
		i -- ;
	}
	// Outing real stack lines
	while(i>0) {
		j = 0 ;
		while( (n<max)&&(rowfmt[j]!=0) ) {
			while( (n<max)&&(rowfmt[j]!=0) ) {
				if(rowfmt[j]=='%') break ;
				out[n] = rowfmt[j] ;
				j ++ ;
				n ++ ;
			}
			if(n==max) {
				out[n] = 0 ;
				return n ;
			}
			if(rowfmt[j]==0) break ;
			j ++ ;
			l = 0 ;
			if(rowfmt[j]==0) break ;
			if( (rowfmt[j]<='0')||(rowfmt[j]>'9') ) {
				fil = rowfmt[j] ;
				j ++ ;
			}
			else fil = ' ' ;
			if(rowfmt[j]==0) break ;
			if( (rowfmt[j]>='0')&&(rowfmt[j]<='9') ) {
				j += _scaninteger(&l,rowfmt+j,10,10) ;
			}
			if(rowfmt[j]==0) break ;
			switch(rowfmt[j]) {
				case 'n':
					if(l==0) l = 4 ;
					if(n+l<max) {
						_printinteger(i,out+n,l,10) ;
						n += l ;
					}
					break ;
				case 's':
					if(l==0) l = 20 ;
					if(n+l<max) {
						r = ObjectOut(stk->data[stk->dpth-i],NULL,l) ;
						if(r==-1) {
							return -1 ;
						}
						if(r>l) r = l ;
						else {
							u = l-r ;
							while(u>0) {
								u -- ;
								out[n] = fil ;
								n ++ ;
							}
						}
						r = ObjectOut(stk->data[stk->dpth-i],out+n,r) ;
						if(r==-1) {
							return -1 ;
						}
						n += r ;
					}
					break ;
				default:
					;
			}
			j ++ ;
		}
		if(n<max) {
			out[n] = '\n' ;
			n ++ ;
		}
		else {
			out[n] = 0 ;
			return n ;
		}
		i -- ;
	}
	j = 0 ;
	while(n<max) {
		if(tail[j]==0) break ;
		out[n] = tail[j] ;
		j ++ ;
		n ++ ;
	}
	if(n<max) {
		out[n] = '\n' ;
		n ++ ;
	}
	out[n] = 0 ;
	return n ;
}


unsigned long StackStat(Stack stk) {
	printf("Stack [%p]\n",stk) ;
	if(stk==NULL) {
		printf("    Buffer size: N/A\n") ;
		printf("    Stack depth: N/A\n") ;
		printf("    Chunk width: N/A\n") ;
		printf("    Buffer address: N/A\n") ;
		return -1 ;
	}
	printf("    Buffer size: %lu\n",stk->size) ;
	printf("    Stack depth: %lu\n",stk->dpth) ;
	printf("    Chunk width: %lu\n",stk->chnk) ;
	printf("    Buffer address: [%p]\n",stk->data) ;
	return 0 ;
}





//////////////////////////////////////////////////////////////////////////////////////////////////
/////
/////  Variables domain definition
/////


// #define NAMELEN 128 ;
//
// struct S_NameSpace {
//	unsigned long size ;
//	unsigned long chnk ;
//	unsigned long dpth ;
//	char * names[NAMELEN+1] ;
//	Object conts ;
// }
//
// typedef struct S_NameSpace * NameSpace ;



NameSpace NameSpaceAlloc(unsigned long chksz) {
	NameSpace nms ;
	nms = (NameSpace) malloc(sizeof(struct S_NameSpace)) ;
	if(nms==NULL) {
		DEBUG("NameSpaceAlloc",0,NULL) ;
		return NULL ;
	}
	nms->chnk = (chksz>0)?chksz:DEFCHNKSIZE ;
	nms->size = 0 ;
	nms->dpth = 0 ;
	nms->names = NULL ;
	nms->conts = NULL ;
	return nms ;
}


void NameSpaceFree(NameSpace nms) {
	register unsigned long i ;
	if(nms!=NULL) {
		if(nms->names!=NULL) {
			i = 0 ;
			while(i<nms->dpth) {
				free(nms->names[i]) ;
				i++ ;
			}
			free(nms->names) ;
			nms->names = NULL ;
		}
		if(nms->conts!=NULL) {
			i = 0 ;
			while(i<nms->dpth) {
				ObjectDrop(nms->conts[i]) ;
				ObjectFree(nms->conts[i]) ;
				i++ ;
			}
			free(nms->conts) ;
			nms->conts = NULL ;
		}
		free(nms) ;
	}
}


Object NameSpaceGet(NameSpace nms, char * name) {
	unsigned long p = 0 ;
	if(nms==NULL) {
		errno = EINVAL ;
		DEBUG("NameSpaceGet",2,"NULL NameSpace") ;
		return NULL ;
	}
	if(name==NULL) {
		errno = EINVAL ;
		DEBUG("NameSpaceGet",2,"NULL name") ;
		return NULL ;
	}
	// p will be filled with the position of the first greater key
	if(_keysearch(name,nms->names,nms->dpth,NAMELEN,&p)!=-1) return nms->conts[p] ;
	return NULL ;
}


Object NameSpaceSet(NameSpace nms, char * name, Object obj) {
	register unsigned long i = 0 ;
	unsigned long p = 0 ;
	char * * newc ;
	Object * newo ;
	Object oldo ;
	char * oldc ;
	if(nms==NULL) {
		errno = EINVAL ;
		DEBUG("NameSpaceSet",2,"NULL NameSpace") ;
		return NULL ;
	}
	if(name==NULL) {
		errno = EINVAL ;
		DEBUG("NameSpaceSet",2,"NULL name") ;
		return NULL ;
	}
	if(obj==NULL) {
		errno = EINVAL ;
		DEBUG("NameSpaceSet",2,"NULL Object") ;
		return NULL ;
	}
	// p will be filled with the position of the first greater key
	if(_keysearch(name,nms->names,nms->dpth,NAMELEN,&p)!=-1) {
		oldo = nms->conts[p] ;
		ObjectDrop(oldo) ;
		ObjectDup(obj) ;
		nms->conts[p] = obj ;
		return oldo ;
	}
	if(nms->chnk==0) {
		errno = EACCES ;
		DEBUG("NameSpaceSet",3,"Freezed NameSpace") ;
		return NULL ;
	}
	if(nms->dpth==nms->size) {
		newc = (char * *) realloc(nms->names,(nms->size+nms->chnk)*sizeof(char *)) ;
		if(newc==NULL) {
			DEBUG("NameSpaceSet",0,NULL) ;
			return NULL ;
		}
		newo = (Object *) realloc(nms->conts,(nms->size+nms->chnk)*sizeof(struct S_Object)) ;
		if(newo==NULL) {
			newc = (char * *) realloc(newc,nms->size*sizeof(char *)) ;
			if(newc!=NULL) nms->names = newc ;
			DEBUG("NameSpaceSet",0,NULL) ;
			return NULL ;
		}
		nms->names = newc ;
		nms->conts = newo ;
		i = 0 ;
		while(i<nms->chnk) {
			nms->names[nms->size+i] = (char *) malloc(NAMELEN+1) ;
			if(nms->names[nms->size+i]==NULL) {
				i -- ;
				while(i>0) {
					free(nms->names[nms->size+i]) ;
					i -- ;
				}
				newc = (char * *) realloc(newc,nms->size*sizeof(char *)) ;
				if(newc!=NULL) nms->names = newc ;
				newo = (Object *) realloc(newo,nms->size*sizeof(struct S_Object)) ;
				if(newo!=NULL) nms->conts = newo ;
				DEBUG("NameSpaceSet",0,NULL) ;
				return NULL ;
			}
			nms->names[nms->size+i][NAMELEN] = 0 ;
			i ++ ;
		}
		nms->size += nms->chnk ;
	}
	i = nms->dpth ;
	oldc = nms->names[i] ;
	while(i>p) {
		nms->names[i] = nms->names[i-1] ;
		nms->conts[i] = nms->conts[i-1] ;
		i -- ;
	}
	nms->names[i] = oldc ; // Recycling NAMELEN+1 allocated chars 
	i = 0 ;
	while(i<NAMELEN) {
		if(name[i]==0) break ;
		nms->names[p][i] = name[i] ;
		i ++ ;
	}
	while(i<NAMELEN) {
		nms->names[p][i] = 0 ;
		i ++ ;
	}
	ObjectDup(obj) ;
	nms->conts[p] = obj ;
	nms->dpth ++ ;
	return obj ;
}


Object NameSpaceDel(NameSpace nms, char * name) {
	register unsigned long i = 0 ;
	unsigned long p = 0 ;
	char * * newc ;
	Object * newo ;
	Object oldo ;
	char * oldc ;
	if(nms==NULL) {
		errno = EINVAL ;
		DEBUG("NameSpaceDel",2,"NULL NameSpace") ;
		return NULL ;
	}
	if(name==NULL) {
		errno = EINVAL ;
		DEBUG("NameSpaceDel",2,"NULL name") ;
		return NULL ;
	}
	if(nms->chnk==0) {
		errno = EACCES ;
		DEBUG("NameSpaceDel",3,"Freezed NameSpace") ;
		return NULL ;
	}
	// p will be filled with the position of the first greater key
	if(_keysearch(name,nms->names,nms->dpth,NAMELEN,&p)!=-1) {
		oldo = nms->conts[p] ;
		ObjectDrop(oldo) ;
		nms->dpth -- ;
		i = p ;
		oldc = nms->names[i] ;
		while(i<nms->dpth) {
			nms->names[i] = nms->names[i+1] ;
			nms->conts[i] = nms->conts[i+1] ;
			i ++ ;
		}
		nms->names[i] = oldc ; // Recycling NAMELEN+1 allocated chars 
		if(nms->dpth+2*nms->chnk<nms->size) {
			nms->size -= nms->chnk ;
			i = 0 ;
			while(i<nms->chnk) {
				free(nms->names[nms->size+i]) ;
				i ++ ;
			}
			newc = (char * *) realloc(nms->names,nms->size*sizeof(char *)) ;
			if(newc==NULL) { // Non-fatal(!) error: can't disallocate
				DEBUG("NameSpaceDel",0,"Disallocation not performed") ;
				return oldo ;
			}
			newo = (Object *) realloc(nms->conts,nms->size*sizeof(struct S_Object)) ;
			if(newo==NULL) { // Non-fatal(!) error: can't disallocate
				DEBUG("NameSpaceDel",0,"Disallocation not performed") ;
				return oldo ;
			}
			nms->names = newc ;
			nms->conts = newo ;
		}
		return oldo ;
	}
	return NULL ;
}


List NameSpaceList(NameSpace nms) {
	List lst = NULL ;
	Strg str = NULL ;
	Object obj = NULL ;
	unsigned long i = 0 ;
	lst = ListAlloc() ;
	while(i<nms->dpth) {
		str = StrgNew(nms->names[i]) ;
		if(str==NULL) {
			ListFree(lst) ;
			return NULL ;
		}
		obj = StrgWrap(str) ;
		if(obj==NULL) {
			ListFree(lst) ;
			return NULL ;
		}
		if(ListPush(lst,obj)==NULL) {
			ListFree(lst) ;
			return NULL ;
		}
		i ++ ;
	}
	return lst ;
}


unsigned long NameSpaceOut(NameSpace nms, char * out, unsigned long max, char * head, char * rowfmt, char * tail) {
	register unsigned long n = 0 ;
	register unsigned long i = 0 ;
	register unsigned long j ;
	register unsigned long u ;
	register unsigned long r ;
	unsigned long l ;
	char fil = ' ' ;  //Fill char


	if(nms==NULL) {
		errno = EINVAL ;
		DEBUG("NameSpaceOut",2,"NULL NameSpace") ;
		return -1 ;
	}

	if(out==NULL) {
		j = 0 ;
		while(rowfmt[j]!=0) {
			u = j ;
			while(rowfmt[j]!=0) {
				if(rowfmt[j]=='%') break ;
				j ++ ;
				n ++ ;
			}
			if(rowfmt[j]==0) break ;
			j ++ ;
			l = 0 ;
			if(rowfmt[j]==0) break ;
			if( (rowfmt[j]<='0')||(rowfmt[j]>'9') ) {
				j ++ ;
			}
			if(rowfmt[j]==0) break ;
			if( (rowfmt[j]>='0')&&(rowfmt[j]<='9') ) {
				j += _scaninteger(&l,rowfmt+j,10,10) ;
			}
			if(rowfmt[j]==0) break ;
			switch(rowfmt[j]) {
				case 'n':
					if(l==0) l = 4 ;
					n += l ;
					break ;
				case 's':
					if(l==0) l = 20 ;
					n += l ;
					break ;
				default:
					;
			}
			j ++ ;
		}
		n *= nms->dpth ;
		j = 0 ;
		while(head[j]!=0) {
			if(head[j]==0) break ;
			j ++ ;
			n ++ ;
		}
		j = 0 ;
		while(tail[j]!=0) {
			j ++ ;
			n ++ ;
		}
		return n ;
	}

	//First line
	j = 0 ;
	while(n<max) {
		if(head[j]==0) break ;
		out[n] = head[j] ;
		j ++ ;
		n ++ ;
	}
	if(n<max) {
		out[n] = '\n' ;
		n ++ ;
	}
	else {
		out[n] = 0 ;
		return n ;
	}
	// Outing variables
	while(i<nms->dpth) {
		// One var line more...
		j = 0 ;
		while( (n<max)&&(rowfmt[j]!=0) ) {
			while( (n<max)&&(rowfmt[j]!=0) ) {
				if(rowfmt[j]=='%') break ;
				out[n] = rowfmt[j] ;
				j ++ ;
				n ++ ;
			}
			if(n==max) {
				out[n] = 0 ;
				return n ;
			}
			if(rowfmt[j]==0) break ;
			j ++ ;
			l = 0 ;
			if(rowfmt[j]==0) break ;
			if( (rowfmt[j]<='0')||(rowfmt[j]>'9') ) {
				fil = rowfmt[j] ;
				j ++ ;
			}
			else fil = ' ' ;
			if(rowfmt[j]==0) break ;
			if( (rowfmt[j]>='0')&&(rowfmt[j]<='9') ) {
				j += _scaninteger(&l,rowfmt+j,10,10) ;
			}
			if(rowfmt[j]==0) break ;
			switch(rowfmt[j]) {
				case 'n':
					if(l==0) l = 4 ;
					if(n+l<max) {
						r = 0 ;
						while(r<l) {
							if(nms->names[i][r]==0) break ;
							r ++ ;
						}
						u = 0 ;
						while(u<(l-r)) {
							out[n] = fil ;
							u ++ ;
							n ++ ;
						}
						u = 0 ;
						while(u<r) {
							if(nms->names[i][u]==0) break ;
							out[n] = nms->names[i][u] ;
							u ++ ;
							n ++ ;
						}
					}
					break ;
				case 's':
					if(l==0) l = 20 ;
					if(n+l<max) {
						r = ObjectOut(nms->conts[i],NULL,l) ;
						if(r==-1) {
							return -1 ;
						}
						if(r>l) r = l ;
						else {
							u = l-r ;
							while(u>0) {
								u -- ;
								out[n] = fil ;
								n ++ ;
							}
						}
						r = ObjectOut(nms->conts[i],out+n,r) ;
						if(r==-1) {
							return -1 ;
						}
						n += r ;
					}
					break ;
				default:
					;
			}
			j ++ ;
		}
		if(n<max) {
			out[n] = '\n' ;
			n ++ ;
		}
		else {
			out[n] = 0 ;
			return n ;
		}
		i ++ ;
	}
	//Lsat line
	j = 0 ;
	while(n<max) {
		if(tail[j]==0) break ;
		out[n] = tail[j] ;
		j ++ ;
		n ++ ;
	}
	if(n<max) {
		out[n] = '\n' ;
		n ++ ;
	}
	out[n] = 0 ;
	return n ;
}


unsigned long NameSpaceStat(NameSpace nms) {
	printf("NameSpace [%p]\n",nms) ;
	if(nms==NULL) {
		printf("    Buffer size : N/A\n") ;
		printf("    NameSpace depth: N/A\n") ;
		printf("    Chunk width : N/A\n") ;
		printf("    Name tab address: N/A\n") ;
		printf("    Objc tab address: N/A\n") ;
		return -1 ;
	}
	printf("    Buffer size : %lu\n",nms->size) ;
	printf("    NameSpace depth: %lu\n",nms->dpth) ;
	printf("    Chunk width : %lu\n",nms->chnk) ;
	printf("    Name tab address: [%p]\n",nms->names) ;
	printf("    Objc tab address: [%p]\n",nms->conts) ;
	return 0 ;
}



